import numpy as np
import matplotlib.pyplot as plt
import ngsolve
import sympy
import pickle

from hydrodynamics import Hydrodynamics, select_model_options
import boundary_fitted_coordinates
from spatial_parameter import SpatialParameter
from TruncationBasis import eigbasis_constantAv, harmonic_time_basis
import mesh_functions
from geometry.geometries import *
from geometry.create_geometry import parametric_geometry, WALLDOWN, WALLUP, RIVER, SEA, BOUNDARY_DICT


# STEP 1: Create geometry and mesh ====================================================================================

# set shape parameters

L = 10e3 # length of estuary in m
B = 3e3 # width of estuary in m

geometrycurves = parametric_rectangle(B, L)
maxh_global = 0.95e3

boundary_maxh_dict = {WALLDOWN: [maxh_global], WALLUP: [maxh_global], RIVER: [maxh_global], SEA: [maxh_global]} # set element size for each partition: allows for variable mesh sizes!
boundary_parameter_partition_dict = {WALLDOWN: [0,1], WALLUP: [0,1], RIVER: [0,1], SEA: [0,1]}  # partition boundary into segments, assuming it starts at 0 and ends at 1

geometry = parametric_geometry(geometrycurves, boundary_parameter_partition_dict, boundary_maxh_dict)

# create mesh

mesh = ngsolve.Mesh(geometry.GenerateMesh(maxh=maxh_global))


# STEP 2: Set parameters =======================================================================================

sem_order = 2
M = 1 # amount of vertical basis functions to be projected on
imax = 1 # amount of tidal constituents in addition to subtidal flow


# constant physical parameters

g = 9.81 # gravitational acceleration
f = 1e-4
Av = 0.01 # vertical eddy viscosity
sigma = 2 / 89428.32720 # M2-frequency (not angular) from Table 3.2 in Gerkema (2019)


# spatially varying physical parameters

bfc = boundary_fitted_coordinates.generate_bfc(mesh, sem_order, "diffusion", alpha=1) # generate coordinates that follow the boundary of the estuary

def H(xi, eta): # Gaussian bathymetry
    return 2 + 10 * sympy.exp(-6 * eta**2)
    

def rho(xi, eta): # water density following the hyperbolic tangent profile from Talke et al. (2009)
    return 0

def R(xi, eta):
    return 0 * xi

H_sp = SpatialParameter(H, bfc)
rho_sp = SpatialParameter(rho, bfc)
R_sp = SpatialParameter(R, bfc)

# STEP 3: Define expansion bases ============================================================================

vertical_basis = eigbasis_constantAv
time_basis = harmonic_time_basis(sigma)


# STEP 4 : Create Hydrodynamics object ======================================================================

# set options

model_options = select_model_options(bed_bc='no_slip',
                                     surface_in_sigma=False,
                                     veddy_viscosity_assumption='constant', 
                                     density='depth-independent',
                                     advection_epsilon=0,
                                     advection_influence_matrix=np.array([[True, True],
                                                                          [False, False]]),
                                     x_scaling=1.,
                                     y_scaling=1.)

# create object

hydro = Hydrodynamics(mesh, model_options, imax, M, sem_order, time_basis, vertical_basis)
hydro.set_constant_physical_parameters(Av=Av, sigma=sigma, g=g, f=f)
hydro.set_spatial_physical_parameters(H=H_sp, density=rho_sp, R=R_sp)


# refine mesh based on bathymetry gradient

hydro.hrefine(1.2, 2)

# set tidal forcing

tide_amplitude_list = [0, 1] # per constituent
tide_phase_list = [0, 0]
hydro.set_seaward_boundary_condition(tide_amplitude_list, tide_phase_list)

# set river discharge (only stationary discharges work right now)

discharge_amplitude_list = [0, 0]
discharge_phase_list = [0, 0]
hydro.set_riverine_boundary_condition(discharge_amplitude_list, discharge_phase_list, is_constant=True)


# STEP 5: Solve the equations ===============================================================================

advection_epsilon_list = [0] # this is a list to allow for homology methods in the Newton method; the solution procedure could really use a rewrite as well; this is not so intuitive

hydro.solve(advection_epsilon_list, skip_nonlinear=True, maxits=1, tol=1e-7, method='pardiso')


mesh.ngmesh.SetGeometry(None)
with open('mesh.pkl', 'wb') as meshfile:
    pickle.dump(mesh, meshfile)

with open('solution.pkl', 'wb') as solfile:
    pickle.dump(hydro.solution_gf, solfile)


with open('mesh.pkl', 'rb') as meshfileload:
    newmesh = pickle.load(meshfileload)

with open('solution.pkl', 'rb') as solfileload:
    newsol = pickle.load(solfileload)


mesh_functions.plot_CF_colormap(hydro.solution_gf.components[1], mesh, refinement_level=3, show_mesh=False, title='Before saving')
mesh_functions.plot_CF_colormap(newsol.components[1], newmesh, refinement_level=3, show_mesh=False, title='After save/load')

plt.show()











